#   fifty20.py  Flytec 5020/5030/6020/6030 and Brauniger Galileo/Competino/Compeo/Competino+/Compeo+ functions
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import logging
import re
import struct

from base import FlightRecorderBase
from common import CTR, CTRPoint, Track, add_igc_filenames, simplerepr
from errors import NotAvailableError, ProtocolError
import nmea
from utc import UTC
from waypoint import Waypoint


logger = logging.getLogger(__name__)


MANUFACTURER = {}
for model in '5020 5030 6020 6030'.split():
    MANUFACTURER[model] = 'Flytec'
for model in 'COMPEO COMPEO+ COMPETINO COMPETINO+ GALILEO'.split():
    MANUFACTURER[model] = 'Brauniger'

MEMORY_MAP = {
        'glider_id': (224, '16s', str),
        'glider_type': (192, '16s', str),
        'pilot_name': (0, '16s', str),
        'recording_interval': (97, 'B', int)}

XON = '\021'
XOFF = '\023'

PBRANS_RE = re.compile(r'\APBRANS,(\d+)\Z')
PBRCTR_RE1 = re.compile(r'\APBRCTR,(\d+),0+,([^,]*),(\d+)\Z')
PBRCTR_RE2 = re.compile(r'\APBRCTR,(\d+),0*1,([^,]*)\Z')
PBRCTR_RE3 = re.compile(r'\APBRCTR,(\d+),(\d+),([CPTXZ]),(\d{2})(\d{2}\.\d{3}),([NS]),(\d{3})(\d{2}\.\d{3}),([EW])')
PBRCTR_PX_RE = re.compile(r'\APBRCTR,\d+,\d+,[PX],\d{4}\.\d{3},[NS],\d{5}\.\d{3},[EW]\Z')
PBRCTR_C_RE = re.compile(r'\APBRCTR,\d+,\d+,C,\d{4}\.\d{3},[NS],\d{5}\.\d{3},[EW],(\d+)\Z')
PBRCTR_TZ_RE = re.compile(r'\APBRCTR,\d+,\d+,[TZ],\d{4}\.\d{3},[NS],\d{5}\.\d{3},[EW],([+\-])\Z')
PBRCTRI_RE = re.compile(r'\APBRCTRI,(\d+),(\d+),(\d+)\Z')
PBRMEMR_RE = re.compile(r'\APBRMEMR,([0-9A-F]+),([0-9A-F]+(?:,[0-9A-F]+)*)\Z')
PBRRTS_RE1 = re.compile(r'\APBRRTS,(\d+),(\d+),0+,(.*)\Z')
PBRRTS_RE2 = re.compile(r'\APBRRTS,(\d+),(\d+),(\d+),([^,]*),(.*?)\Z')
PBRSNP_RE = re.compile(r'\APBRSNP,([^,]*),([^,]*),([^,]*),([^,]*)\Z')
PBRTL_RE = re.compile(r'\APBRTL,(\d+),(\d+),(\d+).(\d+).(\d+),(\d+):(\d+):(\d+),(\d+):(\d+):(\d+)\Z')
PBRTLE_RE = re.compile(r'\APBRTLE,(\d+),(\d+),(\d+).(\d+).(\d+),(\d+):(\d+):(\d+),(\d+):(\d+):(\d+),(\d+),(\d+),(\d+),(\d+)')
PBRWPS_RE = re.compile(r'\APBRWPS,(\d{2})(\d{2}\.\d{3}),([NS]),(\d{3})(\d{2}\.\d{3}),([EW]),([^,]*),([^,]*),(\d+)\Z')


class CTRI(object):

    def __init__(self, actual, maximum, free):
        self.actual = actual
        self.maximum = maximum
        self.free = free

    __repr__ = simplerepr

    __repr__ = simplerepr


class Route(object):

    def __init__(self, index, name, routepoints):
        self.index = index
        self.name = name
        self.routepoints = routepoints

    __repr__ = simplerepr


class Routepoint(object):

    def __init__(self, short_name, long_name):
        self.short_name = short_name
        self.long_name = long_name

    __repr__ = simplerepr


class SNP(object):

    def __init__(self, model, pilot_name, serial_number, software_version):
        self.model = model
        self.pilot_name = pilot_name.strip()
        self.serial_number = int(serial_number)
        self.software_version = software_version

    __repr__ = simplerepr


class Fifty20(FlightRecorderBase):

    SUPPORTED_MODELS = '5020 5030 6020 6030 COMPEO COMPEO+ COMPETINO COMPETINO+ GALILEO'.split()

    def __init__(self, io, line=None):
        self.io = io
        self.buffer = ''
        self._snp = SNP(*PBRSNP_RE.match(line[1:-1].decode('nmea_sentence')).groups()) if line else None
        self._tracks = None
        self._waypoints = None
        self.waypoint_precision = 1

    def readline(self, timeout=1):
        if self.buffer == '':
            self.buffer = self.io.read(timeout)
        if self.buffer[0] == XON:
            self.buffer = self.buffer[1:]
            logger.debug('read XON')
            return XON
        elif self.buffer[0] == XOFF:
            self.buffer = self.buffer[1:]
            logger.debug('read XOFF')
            return XOFF
        else:
            line = ''
            while True:
		index = self.buffer.find('\n')
		if index == -1:
                    line += self.buffer
                    self.buffer = self.io.read(timeout)
		else:
                    line += self.buffer[:index + 1]
                    self.buffer = self.buffer[index + 1:]
                    logger.info('readline %r' % line)
                    return line

    def write(self, line):
        logger.info('write %r' % line)
        self.io.write(line)

    def ieach(self, command, re=None, timeout=1):
        try:
            self.write(command.encode('nmea_sentence'))
            if self.readline(timeout) != XOFF:
                raise ProtocolError
            while True:
		line = self.readline(timeout)
		if line == XON:
                    break
		elif re is None:
                    yield line
		else:
                    m = re.match(line.decode('nmea_sentence'))
                    if m is None:
                        raise ProtocolError(line)
                    yield m
        except:
            self.io.flush()
            raise

    def none(self, command, timeout=1):
        for m in self.ieach(command, None, timeout):
            raise ProtocolError(m)

    def one(self, command, re=None, timeout=1):
        result = None
        for m in self.ieach(command, re, timeout):
            if not result is None:
                raise ProtocolError(m)
            result = m
        return result

    def pbrconf(self):
        self.none('PBRCONF,', None)
        self._snp = None

    def ipbrctr(self):
        for l in self.ieach('PBRCTR,'):
            l = l.decode('nmea_sentence')
            m = PBRCTR_RE1.match(l)
            if m:
                count, name, warning_distance = int(m.group(1)), m.group(2), int(m.group(3))
                remark = None
                ctrpoints = []
                continue
            m = PBRCTR_RE2.match(l)
            if m:
                if int(m.group(1)) != count:
                    raise ProtocolError('count mismatch')
                remark = m.group(2)
                continue
            m = PBRCTR_RE3.match(l)
            if m:
                if int(m.group(1)) != count:
                    raise ProtocolError('count mismatch')
                index, type = int(m.group(2)), m.group(3)
                lat = int(m.group(4)) + float(m.group(5)) / 60
                if m.group(6) == 'S':
                    lat *= -1
                lon = int(m.group(7)) + float(m.group(8)) / 60
                if m.group(9) == 'W':
                    lon *= -1
                radius, clockwise = None, None
                if type in ('P', 'X'):
                    m = PBRCTR_PX_RE.match(l)
                    if not m:
                        raise ProtocolError(l)
                elif type == 'C':
                    m = PBRCTR_C_RE.match(l)
                    if not m:
                        raise ProtocolError(l)
                    radius = int(m.group(1))
                elif type in ('T', 'Z'):
                    m = PBRCTR_TZ_RE.match(l)
                    if not m:
                        raise ProtocolError(l)
                    clockwise = m.group(1) == '+'
                else:
                    raise ProtocolError(l)
                ctrpoint = CTRPoint(type, lat, lon, radius, clockwise)
                ctrpoints.append(ctrpoint)
                if index == count - 1:
                    yield CTR(name, warning_distance, remark, ctrpoints)
                continue
            m = PBRANS_RE.match(l)
            if m:
                status = int(m.group(1))
                if status != 1:
                    raise RuntimeError(l)  # FIXME
                continue
            logger.warning('unexpected response %r', l)

    def pbrctri(self):
        return CTRI(*map(int, self.one('PBRCTRI', PBRCTRI_RE).groups()))

    def pbrctrw(self, ctr, warning_distance):
        n = 2 + len(ctr.ctrpoints)
        self.none('PBRCTRW,%03d,000,%s,%04d' % (n, ctr.name[:17].ljust(17), ctr.warning_distance or warning_distance))
        self.none('PBRCTRW,%03d,001,%s' % (n, ctr.remark[:17].ljust(17)))
        for i, ctrpoint in enumerate(ctr.ctrpoints):
            command = 'PBRCTRW,%03d,%03d,%s,%02d%06.3f,%s,%03d%06.3f,%s' % (
                    n,
                    i + 2,
                    ctrpoint.type,
                    abs(60 * ctrpoint.lat) / 60,
                    abs(60 * ctrpoint.lat) % 60,
                    'S' if ctrpoint.lat < 0 else 'N',
                    abs(60 * ctrpoint.lon) / 60,
                    abs(60 * ctrpoint.lon) % 60,
                    'W' if ctrpoint.lon < 0 else 'E')
            if ctrpoint.type == 'C':
                command += ',%04d' % (ctrpoint.radius,)
            elif ctrpoint.type in ('T', 'Z'):
                command += ',%s' % ('+' if ctrpoint.clockwise else '-',)
            if i == len(ctr.ctrpoints) - 1:
                status = int(self.one(command, PBRANS_RE).group(1))
                if status != 1:
                    raise ProtocolError('status == %d' % (status,))
            else:
                self.none(command)

    def ipbrigc(self):
        return self.ieach('PBRIGC,')

    def pbrmemr(self, address, length):
        result = []
        first, last = address, address + length
        while first < last:
            m = self.one('PBRMEMR,%04X' % first, PBRMEMR_RE)
            if int(m.group(1), 16) != first:
                raise ProtocolError('address mismatch')
            data = list(int(i, 16) for i in m.group(2).split(','))
            result.extend(data)
            first += len(data)
        return result[:length]

    def pbrmemw(self, address, value):
        while value:
            chunk = value[:8]
            m = self.one('PBRMEMW,%04X,%d%s%s' % (address, len(chunk), ''.join(',%02X' % ord(c) for c in chunk), ',' * (8 - len(chunk))), PBRMEMR_RE)
            if int(m.group(1), 16) != address:
                raise ProtocolError('address mismatch')
            if not ''.join(chr(int(i, 16)) for i in m.group(2).split(',')).startswith(chunk):
                raise ProtocolError('readback mismatch')
            address += len(chunk)
            value = value[len(chunk):]

    def ipbrrts(self):
        for l in self.ieach('PBRRTS,'):
            l = l.decode('nmea_sentence')
            m = PBRRTS_RE1.match(l)
            if m:
		index, count, name = int(m.group(1)), int(m.group(2)), m.group(3)
		if count == 1:
                    yield Route(index, name, [])
		else:
                    routepoints = []
            else:
		m = PBRRTS_RE2.match(l)
		if m:
                    index, count, routepoint_index = (int(i) for i in m.groups()[0:3])
                    routepoint_short_name = m.group(4)
                    routepoint_long_name = m.group(5)
                    routepoints.append(Routepoint(routepoint_short_name, routepoint_long_name))
                    if routepoint_index == count - 1:
                        yield Route(index, name, routepoints)
		else:
                    raise ProtocolError(m)

    def pbrrts(self):
        return list(self.ipbrrts())

    def pbrsnp(self):
        return SNP(*self.one('PBRSNP,', PBRSNP_RE, 0.2).groups())

    def pbrtl(self):
        tracks = []
        def igc_lambda(self, index):
            return lambda: self.ipbrtr(index)
        for m in self.ieach('PBRTL,', PBRTL_RE, 0.5):
            index = int(m.group(2))
            day, month, year, hour, minute, second = (int(i) for i in m.groups()[2:8])
            hours, minutes, seconds = (int(i) for i in m.groups()[8:11])
            tracks.append(Track(
                count=int(m.group(1)),
                index=index,
                datetime=datetime.datetime(year + 2000, month, day, hour, minute, second, tzinfo=UTC()),
                duration=datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds),
                _igc_lambda=igc_lambda(self, index)))
        return add_igc_filenames(tracks, self.manufacturer[:3].upper(), self.serial_number)

    def pbrtle(self):
        tracks = []
        def igc_lambda(self, index):
            return lambda: self.ipbrtr(index)
        for m in self.ieach('PBRTLE,', PBRTLE_RE, 0.5):
            index = int(m.group(2))
            day, month, year, hour, minute, second = (int(i) for i in m.groups()[2:8])
            hours, minutes, seconds = (int(i) for i in m.groups()[8:11])
            tracks.append(Track(
                count=int(m.group(1)),
                index=index,
                datetime=datetime.datetime(year + 2000, month, day, hour, minute, second, tzinfo=UTC()),
                duration=datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds),
                max_a1=int(m.group(12)),
                max_a2=int(m.group(13)),
                max_a3=int(m.group(14)),
                scan_rate=int(m.group(15)),
                _igc_lambda=igc_lambda(self, index)))
        return add_igc_filenames(tracks, self.manufacturer, self.serial_number)

    def ipbrtr(self, index):
        return self.ieach('PBRTR,%02d' % index)

    def pbrtr(self, index):
        return list(self.ipbrtr(index))

    def pbrwpr(self, waypoint):
        name = waypoint.get_id_name().encode('nmea_characters')[:17].ljust(17)
        self.none('PBRWPR,%02d%06.3f,%s,%03d%06.3f,%s,,%s,%04d' % (
            abs(60 * waypoint.lat) / 60,
            abs(60 * waypoint.lat) % 60,
            'S' if waypoint.lat < 0 else 'N',
            abs(60 * waypoint.lon) / 60,
            abs(60 * waypoint.lon) % 60,
            'W' if waypoint.lon < 0 else 'E',
            name,
            waypoint.alt or 0))
        return name

    def ipbrwps(self):
        for m in self.ieach('PBRWPS,', PBRWPS_RE):
            lat = int(m.group(1)) + float(m.group(2)) / 60
            if m.group(3) == 'S':
                lat *= -1
            lon = int(m.group(4)) + float(m.group(5)) / 60
            if m.group(6) == 'W':
                lon *= -1
            id = m.group(7)
            name = m.group(8)
            alt = int(m.group(9))
            yield Waypoint(name, lat, lon, alt, id=id)

    def pbrwps(self):
        return list(self.ipbrwps())

    def pbrwpx(self, name=None):
        if name:
            self.none('PBRWPX,%-17s' % name)
        else:
            # PBRWPX,, is officially documented but very slow and very buggy
            # so, instead, pretend that the command is not available
            #self.none('PBRWPX,,', None)
            raise NotAvailableError

    @property
    def extended_commands(self):
        return self.snp.model in set(('6020', '6030', 'COMPETINO+', 'COMPEO+'))

    @property
    def manufacturer(self):
        return MANUFACTURER[self.snp.model]

    @property
    def model(self):
        return self.snp.model

    @property
    def serial_number(self):
        return self.snp.serial_number

    @property
    def software_version(self):
        return self.snp.software_version

    @property
    def pilot_name(self):
        return self.snp.pilot_name

    @property
    def snp(self):
        if self._snp is None:
            self._snp = self.pbrsnp()
        return self._snp

    def ctri(self):
        return self.pbrctri()

    def ctrs(self):
        return self.ipbrctr()

    def ctr_upload(self, ctr, warning_distance):
        self.pbrctrw(ctr, warning_distance)

    def get(self, key):
        if key not in MEMORY_MAP:
            raise NotAvailableError
        address, format, type = MEMORY_MAP[key]
        value = ''.join(chr(byte) for byte in self.pbrmemr(address, struct.calcsize(format)))
        return struct.unpack(format, value)[0]

    def set(self, key, value, first=True, last=True):
        if key not in MEMORY_MAP:
            raise NotAvailableError
        address, format, type = MEMORY_MAP[key]
        m = re.match(r'(\d+)s\Z', format)
        if m:
            width = int(m.group(1))
            value = value[:width].ljust(width)
        self.pbrmemw(address, struct.pack(format, type(value)))
        if last:
            self.pbrconf()

    def tracks(self):
        if self._tracks is None:
            self._tracks = self.pbrtl()
        return self._tracks

    def waypoints(self):
        return self.ipbrwps()

    def waypoint_remove(self, name=None):
        if name is None:
            for name in [w.name for w in self.ipbrwps()]:
                self.pbrwpx(name)
        else:
            self.pbrwpx(name)

    def waypoint_upload(self, waypoint):
        return self.pbrwpr(waypoint)

    def to_json(self):
        memory = self.pbrmemr(0, 256)
        tracks = list(track.to_json(True) for track in self.tracks())
        waypoints = list(waypoint.to_json() for waypoint in self.waypoints())
        return dict(memory=memory, tracks=tracks, waypoints=waypoints)



#   error.py  Flytec and Brauniger errors
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Error(RuntimeError):

    def __init__(self, msg=None):
        RuntimeError.__init__(self)
        self.msg = msg


class FlashError(Error):
    pass


class NotAvailableError(Error):
    pass


class TimeoutError(Error):
    pass


class ReadError(Error):
    pass


class WriteError(Error):
    pass


class ProtocolError(Error):
    pass



#   Flymaster low-level device functions
#   Copyright (C) 2010  Tom Payne
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import logging
import re
import struct

from base import FlightRecorderBase
from common import Track, add_igc_filenames
from errors import NotAvailableError, ProtocolError, TimeoutError
import nmea
from utc import UTC
from waypoint import Waypoint


logger = logging.getLogger(__name__)


EPOCH = datetime.datetime(2000, 1, 1, 0, 0, 0)
PBRSNP_RE = re.compile(r'PBRSNP,([^,]*),([^,]*),([^,]*),([^,]*),([^,]*),([^,]*)\Z')
PFMCFG_RE = re.compile(r'FMCFG,([A-Z]+):(.*)\Z')
PFMDNL_LST_RE = re.compile(r'PFMLST,(\d+),(\d+),(\d+).(\d+).(\d+),(\d+):(\d+):(\d+),(\d+):(\d+):(\d+)\Z')
PFMSNP_RE = re.compile(r'PFMSNP,([^,]*),([^,]*),([^,]*),([^,]*),([^,]*),([^,]*)\Z')
PFMWPL_RE = re.compile(r'PFMWPL,(\d{3}\.\d{4}),([NS]),(\d{3}\.\d{4}),([EW]),(\d+),([^,]*),([01])\Z')
PFMWPR_RE = re.compile(r'PFMWPR,ACK,([^,]*)\Z')
TRAILING_NULS_RE = re.compile(r'\x00+')



class Packet(object):

    def __init__(self, id, data):
        self.id = id
        self.data = data


class _Struct:

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__dict__)


class SNP(_Struct):

    def __init__(self, model, _1, serial_number, software_version, _4, _5):
        self.model = model
        self.serial_number = int(serial_number)
        self.software_version = software_version


class FlightInformationRecord(_Struct):

    def __init__(self, data):
        fields = struct.unpack('<BBBBI8s15s15s15s', data[:61])
        self.software_version = '%d.%02d' % (fields[0], fields[1])
        self.hardware_version = '%d.%02d' % (fields[2], fields[3])
        self.serial_number = fields[4]
        self.competition_id = TRAILING_NULS_RE.sub('', fields[5])
        self.pilot_name = TRAILING_NULS_RE.sub('', fields[6])
        self.glider_brand = TRAILING_NULS_RE.sub('', fields[7])
        self.glider_model = TRAILING_NULS_RE.sub('', fields[8])


class KeyTrackPositionRecord(_Struct):

    def __init__(self, data):
        fields = struct.unpack('<BiihhI', data)
        self.fix_flag = fields[0]
        self.lat = fields[1]
        self.lon = fields[2]
        self.alt = fields[3]
        self.pressure = fields[4]
        self.dt = EPOCH + datetime.timedelta(seconds=fields[5])


class TrackPositionRecordDelta(_Struct):

    def __init__(self, data):
        fields = struct.unpack('<Bbbbbb', data)
        self.fix_flag = fields[0]
        self.lat_offset = fields[1]
        self.lon_offset = fields[2]
        self.alt_offset = fields[3]
        self.pressure_offset = fields[4]
        self.dt_offset = datetime.timedelta(seconds=fields[5])


class TrackPositionRecordDeltas(list):

    def __init__(self, data):
        i = 0
        while i < len(data):
            self.append(TrackPositionRecordDelta(data[i:i + 6]))
            i += 6


class Flymaster(FlightRecorderBase):

    SUPPORTED_MODELS = 'B1NAV'.split()

    def __init__(self, io, line=None):
        self.io = io
        self._snp = SNP(*PBRSNP_RE.match(line.decode('nmea_sentence')).groups()) if line else None
        self._pfmdnl_lst = None
        self.buffer = ''
        self.waypoint_precision = 15

    def readline(self, timeout):
        result = ''
        while True:
            index = self.buffer.find('\n')
            if index == -1:
                result += self.buffer
                self.buffer = self.io.read(timeout)
            else:
                result += self.buffer[0:index + 1]
                self.buffer = self.buffer[index + 1:]
                logger.info('readline %r' % result)
                return result

    def readpacket(self, timeout):
        while True:
            s = None
            while True:
                if len(self.buffer) >= 2:
                    id = struct.unpack('<H', self.buffer[:2])[0]
                    if id == 0xa3a3:
                        logger.info('readpacket %r' % self.buffer[:2])
                        self.buffer = self.buffer[2:]
                        return Packet(id, None)
                    if len(self.buffer) >= 4:
                        length = ord(self.buffer[2])
                        if len(self.buffer) >= 4 + length:
                            s = self.buffer[:4 + length]
                            self.buffer = self.buffer[4 + length:]
                            break
                self.buffer += self.io.read(timeout)
            logger.info('readpacket %r' % s[:length + 4])
            data = s[3:length + 3]
            checksum = length
            for c in data:
                checksum ^= ord(c)
            if checksum != ord(s[length + 3]):
                self.write('\xb2')
                continue
            self.write('\xb1')
            return Packet(id, data)

    def write(self, line):
        logger.info('write %r' % line)
        self.io.write(line)

    def ieach(self, command, re=None, timeout=1):
        self.write(command.encode('nmea_sentence'))
        while True:
            line = self.readline(timeout)
            if re is None:
                yield line
            else:
                m = re.match(line.decode('nmea_sentence'))
                if m is None:
                    raise ProtocolError(line)
                yield m

    def none(self, command):
        self.write(command.encode('nmea_sentence'))

    def one(self, command, re=None, timeout=1):
        for m in self.ieach(command, re, timeout=timeout):
            return m

    def pfmids(self, civl_id=None, competition_id=None, pilot_name=None):
        self.none('PFMIDS,%s,%s,%s' % tuple(x[:w].ljust(w) if x else '' for x, w in ((civl_id, 7), (competition_id, 7), (pilot_name, 15))))

    def ipfmcfg(self):
        try:
            for m in self.ieach('PFMCFG,', PFMCFG_RE):
                yield m.groups()
        except TimeoutError:
            pass # FIXME

    def pfmcfg(self):
        return dict(self.ipfmcfg())

    def pfmsnp(self):
        return SNP(*self.one('PFMSNP,', PFMSNP_RE).groups())

    def igc_helper(self, packets):
        yield 'AFLYMASTER %s %s\r\n' % (self.model, self.serial_number)
        date, lat, lon, alt, pressure, dt = None, None, None, None, None, None
        for packet in packets:
            if isinstance(packet, FlightInformationRecord):
                yield 'HFPLTPILOT:%s\r\n' % packet.pilot_name
                yield 'HPGTYGLIDERTYPE:%s %s\r\n' % (packet.glider_brand, packet.glider_model)
                yield 'HPCIDCOMPETITIONID:%s\r\n' % packet.competition_id
                yield 'HFRFWFIRMWAREVERSION:%s\r\n' % packet.software_version
                yield 'HFRHWHARDWAREVERSION:%s\r\n' % packet.hardware_version
                yield 'HFFTYFRTYPE:FLYMASTER,%s\r\n' % self.model
            elif isinstance(packet, KeyTrackPositionRecord):
                if packet.dt.date() != date:
                    yield 'HFDTE%s\r\n' % packet.dt.strftime('%d%m%y')
                    date = packet.dt.date()
                lat, lon, alt, pressure, dt = packet.lat, packet.lon, packet.alt, packet.pressure, packet.dt
                yield 'B%s%02d%02d%03d%c%03d%02d%03d%c%c%05d%05d\r\n' % (
                        dt.strftime('%H%M%S'),
                        abs(lat) / 60000, (abs(lat) % 60000) / 1000, abs(lat) % 1000, 'S' if lat < 0 else 'N',
                        abs(lon) / 60000, (abs(lon) % 60000) / 1000, abs(lon) % 1000, 'E' if lon < 0 else 'W',
                        'A' if packet.fix_flag & 0x80 else 'V',
                        Flymaster.pressure_altitude(pressure),
                        alt)
            elif isinstance(packet, TrackPositionRecordDeltas):
                if lat is None:
                    logger.debug('Track position record delta received before key track position record' % packet)
                    continue
                for tprd in packet:
                    lat += tprd.lat_offset
                    lon += tprd.lon_offset
                    alt += tprd.alt_offset
                    pressure += tprd.pressure_offset
                    dt += tprd.dt_offset
                    if dt.date() != date:
                        yield 'HFDTE%s\r\n' % dt.strftime('%d%m%y')
                        date = dt.date()
                    yield 'B%s%02d%02d%03d%c%03d%02d%03d%c%c%05d%05d\r\n' % (
                            dt.strftime('%H%M%S'),
                            abs(lat) / 60000, (abs(lat) % 60000) / 1000, abs(lat) % 1000, 'S' if lat < 0 else 'N',
                            abs(lon) / 60000, (abs(lon) % 60000) / 1000, abs(lon) % 1000, 'E' if lon < 0 else 'W',
                            'A' if tprd.fix_flag & 0x80 else 'V',
                            Flymaster.pressure_altitude(pressure),
                            alt)

    def pfmdnl_lst(self):
        tracks = []
        def igc_lambda(self, dt):
            return lambda: self.igc_helper(self.ipfmdnl(dt))
        for m in self.ieach('PFMDNL,LST,', PFMDNL_LST_RE):
            count, index, day, month, year, hour, minute, second = map(int, m.groups()[:8])
            hours, minutes, seconds = map(int, m.groups()[8:11])
            dt = datetime.datetime(year + 2000, month, day, hour, minute, second, tzinfo=UTC())
            tracks.append(Track(
                index=index,
                datetime=dt,
                duration=datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds),
                _igc_lambda=igc_lambda(self, dt)))
            if index + 1 == count:
                break
        return add_igc_filenames(tracks, 'XFR', self.serial_number)

    def ipfmdnl(self, dt, timeout=1):
        self.write(('PFMDNL,%s,' % dt.strftime('%y%m%d%H%M%S')).encode('nmea_sentence'))
        while True:
            packet = self.readpacket(timeout)
            if packet.id == 0xa0a0:
                yield FlightInformationRecord(packet.data)
            elif packet.id == 0xa1a1:
                yield KeyTrackPositionRecord(packet.data)
            elif packet.id == 0xa2a2:
                yield TrackPositionRecordDeltas(packet.data)
            elif packet.id == 0xa3a3:
                break
            else:
                logger.info('unknown packet type %04X' % packet.id)

    def ipfmwpl(self):
        try:
            for m in self.ieach('PFMWPL,', PFMWPL_RE):
                lat = float(m.group(1))
                if m.group(2) == 'S':
                    lat = -lat
                lon = float(m.group(3))
                if m.group(4) == 'W':
                    lon = -lon
                alt = int(m.group(5))
                name = m.group(6)
                airfield = bool(int(m.group(7)))
                yield Waypoint(name, lat, lon, alt, airfield=airfield)
        except TimeoutError:
            pass # FIXME

    def pfmwpl(self):
        return list(self.ipfmwpl())

    def pfmwpr(self, waypoint):
        name = re.sub(r'[^ 0-9A-Z]+', lambda m: ' ' * len(m.group(0)), waypoint.get_id_name().upper())[:16].ljust(16)
        m = self.one('PFMWPR,%02d%06.3f,%s,%03d%06.3f,%s,,%s,%04d,%d' % (
            abs(60 * waypoint.lat) / 60,
            abs(60 * waypoint.lat) % 60,
            'S' if waypoint.lat < 0 else 'N',
            abs(60 * waypoint.lon) / 60,
            abs(60 * waypoint.lon) % 60,
            'W' if waypoint.lon < 0 else 'E',
            name,
            waypoint.alt or 0,
            waypoint.airfield), PFMWPR_RE)
        if name != m.group(1):
            raise ProtocolError
        return name

    @property
    def manufacturer(self):
        return 'Flymaster'

    @property
    def model(self):
        if self._snp is None:
            self._snp = self.pfmsnp()
        return self._snp.model

    @property
    def software_version(self):
        if self._snp is None:
            self._snp = self.pfmsnp()
        return self._snp.software_version

    @property
    def serial_number(self):
        if self._snp is None:
            self._snp = self.pfmsnp()
        return self._snp.serial_number

    @property
    def pilot_name(self):
        return None

    def set(self, key, value, first=True, last=True):
        if key in ('civl_id', 'competition_id', 'pilot_name'):
            self.pfmids(**{key: value})
        else:
            raise NotAvailableError

    def tracks(self):
        if self._pfmdnl_lst is None:
            self._pfmdnl_lst = self.pfmdnl_lst()
        return self._pfmdnl_lst

    def waypoints(self):
        return self.pfmwpl()

    def waypoint_upload(self, waypoint):
        return self.pfmwpr(waypoint)

    def to_json(self):
        tracks = list(track.to_json() for track in self.tracks())
        waypoints = list(waypoint.to_json() for waypoint in self.waypoints())
        return dict(tracks=tracks, waypoints=waypoints)

    @staticmethod
    def pressure_altitude(pressure):
        return (1.0 - pow(abs((pressure / 10.0) / 1013.25), 0.190284)) * 44307.69



#   utc.py  UTC timezone
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime


class UTC(datetime.tzinfo):

    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)



#   sixty15.py  Flytec 6015 and Brauniger IQ Basic functions
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
from itertools import chain
import logging
import re
import struct
import time

from base import FlightRecorderBase
from common import Track, add_igc_filenames
from errors import FlashError, NotAvailableError, ProtocolError, ReadError
from utc import UTC
from waypoint import Waypoint


logger = logging.getLogger(__name__)


FA_FORMAT = {}
FA_Owner            = 0x00; FA_FORMAT[FA_Owner]            = '16s'
FA_AC_Type          = 0x01; FA_FORMAT[FA_AC_Type]          = '16s'
FA_AC_ID            = 0x02; FA_FORMAT[FA_AC_ID]            = '16s'
FA_Units            = 0x03; FA_FORMAT[FA_Units]            = 'H'
FA_DiverseFlag      = 0x04; FA_FORMAT[FA_DiverseFlag]      = 'H'
FA_FiltTyp          = 0x05; FA_FORMAT[FA_FiltTyp]          = 'B'
FA_Alt1Diff         = 0x06; FA_FORMAT[FA_Alt1Diff]         = 'l'
FA_VarioDigFk       = 0x07; FA_FORMAT[FA_VarioDigFk]       = 'B'
FA_BFreqRise        = 0x08; FA_FORMAT[FA_BFreqRise]        = 'H'
FA_BFreqSink        = 0x09; FA_FORMAT[FA_BFreqSink]        = 'H'
FA_AudioRise        = 0x0a; FA_FORMAT[FA_AudioRise]        = 'h'
FA_AudioSink        = 0x0b; FA_FORMAT[FA_AudioSink]        = 'h'
FA_SinkAlarm        = 0x0c; FA_FORMAT[FA_SinkAlarm]        = 'h'
FA_FreqGain         = 0x0d; FA_FORMAT[FA_FreqGain]         = 'B'
FA_PitchGain        = 0x0e; FA_FORMAT[FA_PitchGain]        = 'B'
FA_MaxRiseRejection = 0x0f; FA_FORMAT[FA_MaxRiseRejection] = 'H'
FA_VarioMinMaxFk    = 0x10; FA_FORMAT[FA_VarioMinMaxFk]    = 'B'
FA_RecIntervall     = 0x11; FA_FORMAT[FA_RecIntervall]     = 'B'
FA_AudioVolume      = 0x12; FA_FORMAT[FA_AudioVolume]      = 'B'
FA_UTC_Offset       = 0x13; FA_FORMAT[FA_UTC_Offset]       = 'b'
FA_PressOffset      = 0x14; FA_FORMAT[FA_PressOffset]      = 'l'
FA_ThermThreshold   = 0x15; FA_FORMAT[FA_ThermThreshold]   = 'h'
FA_PowerOffTime     = 0x16; FA_FORMAT[FA_PowerOffTime]     = 'B'
FA_StallSpeed       = 0x1a; FA_FORMAT[FA_StallSpeed]       = 'H'
FA_WindWheelGain    = 0x1c; FA_FORMAT[FA_WindWheelGain]    = 'B'
FA_PreThermalThr    = 0x22; FA_FORMAT[FA_PreThermalThr]    = 'h'

FA_MAP = {
        'pilot_name': FA_Owner,
        'glider_type': FA_AC_Type,
        'glider_id': FA_AC_ID,
        'utc_offset': FA_UTC_Offset}

PA_FORMAT = {}
PA_DeviceNr         = 0x00; PA_FORMAT[PA_DeviceNr]         = 'I'
PA_DeviceTyp        = 0x01; PA_FORMAT[PA_DeviceTyp]        = 'B'
PA_SoftVers         = 0x02; PA_FORMAT[PA_SoftVers]         = 'H'
PA_KalibType        = 0x03; PA_FORMAT[PA_KalibType]        = 'B'
PA_Filt1_K          = 0x04; PA_FORMAT[PA_Filt1_K]          = '4B'
PA_Filt2_K          = 0x05; PA_FORMAT[PA_Filt2_K]          = '4B'
PA_Filt4_K          = 0x06; PA_FORMAT[PA_Filt4_K]          = '4B'
PA_AudioHyst        = 0x07; PA_FORMAT[PA_AudioHyst]        = '4B'
PA_AudioRsThrFaktor = 0x08; PA_FORMAT[PA_AudioRsThrFaktor] = '4B'
PA_BattLevel1       = 0x09; PA_FORMAT[PA_BattLevel1]       = '10B'
PA_BattLevel2       = 0x0a; PA_FORMAT[PA_BattLevel2]       = '10B'
PA_BattLevel3       = 0x0b; PA_FORMAT[PA_BattLevel3]       = '10B'
PA_AltiDiff_FLA     = 0x0c; PA_FORMAT[PA_AltiDiff_FLA]     = 'l'
PA_Vario_FLA        = 0x0d; PA_FORMAT[PA_Vario_FLA]        = 'h'
PA_Speed_FLA        = 0x0e; PA_FORMAT[PA_Speed_FLA]        = 'H'
PA_MemoStartDelay   = 0x0f; PA_FORMAT[PA_MemoStartDelay]   = 'B'
PA_Vario_FLE        = 0x10; PA_FORMAT[PA_Vario_FLE]        = 'h'
PA_Speed_FLE        = 0x11; PA_FORMAT[PA_Speed_FLE]        = 'H'


VALID_CHARS = list(chain(
    (0x20, 0x26, 0x28, 0x29),
    xrange(0x2d, 0x3b),
    xrange(0x3c, 0x3f),
    xrange(0x41, 0x5b),
    (0x5f,),
    xrange(0x61, 0x7b)))
INVALID_CHARS_RE = re.compile(r'[^%s]+' % ''.join('\\x%02x' % c for c in VALID_CHARS))


FLASH_RESPONSE_RE = re.compile(r'(S007Flash6015_V1F8|S0056015Wait39|S004Erase_4C|S003Done16|S004Error_32|S005NotReady74)')


class Sixty15(FlightRecorderBase):

    SUPPORTED_MODELS = '6015 IQ-BASIC'.split()

    ATTRIBUTES = {
            'Flytec 6015': {'manufacturer': 'Flytec', 'model': '6015'},
            'IQ-Basic GPS': {'manufacturer': 'Brauniger', 'model': 'IQ-Basic'}}

    def __init__(self, io, line=None):
        self.io = io
        self.buffer = ''
        self._bd = line.rstrip() if line else None
        self._serial_number = None
        self._manufacturer = None
        self._model = None
        self._software_version = None
        self._pilot_name = None
        self._tracks = None
        self._waypoints = None
        self.waypoint_precision = 1

    def readline(self, timeout=1):
        while True:
            index = self.buffer.find('\r\n')
            if index == -1:
                data = self.io.read(timeout)
                if len(data) == 0:
                    raise ReadError
                self.buffer += data
            else:
                line = self.buffer[:index + 2]
                self.buffer = self.buffer[index + 2:]
                logger.info('readline %r' % line)
                return line

    def write(self, line):
        logger.info('write %r' % line)
        self.io.write(line)

    def act1x(self, x, table):
        self.write('ACT_%02X_00\r\n' % x)
        while True:
            line = self.readline()
            if line == ' Done\r\n':
                break
            index, size = map(int, re.split(r'\s*;\s*', line))
            if index not in table:
                logger.warning('field %d not found in table' % index)
            elif struct.calcsize(table[index]) != size:
                logger.error('field %d expected size %d, got %d' % (index, struct.calcsize(table[index]), size))
            else:
                logger.info('field %d size matches' % index)

    def act10(self):
        self.act1x(0x10, FA_FORMAT)

    def act11(self):
        self.act1x(0x11, PA_FORMAT)

    def act20(self):
        self.write('ACT_20_00\r\n')
        line = self.readline(0.5)
        if re.match('\A\s*No\s+Data\s*\r\n\Z', line):
            return []
        tracks = []
        def igc_lambda(self, index):
            return lambda: self.iact21(index)
        while True:
            if line == ' Done\r\n':
                break
            fields = re.split(r'\s*;\s*', line)
            index = int(fields[0])
            year, month, day = (int(x) for x in fields[1].split('.'))
            hour, minute, second = (int(x) for x in fields[2].split(':'))
            hours, minutes, seconds = (int(x) for x in fields[4].split(':'))
            tracks.append(Track(
                    index=index,
                    datetime=datetime.datetime(year + 2000, month, day, hour, minute, second, tzinfo=UTC()),
                    utc_offset=int(fields[3]),
                    duration=datetime.timedelta(seconds=3600 * hours + 60 * minutes + seconds),
                    altitude_offset=int(fields[5]),
                    altitude_max=int(fields[6]),
                    altitude_min=int(fields[7]),
                    vario_max=float(fields[8]),
                    vario_min=float(fields[9]),
                    speed_max=float(fields[10]),
                    pilot_name=fields[11].strip(),
                    glider_type=fields[12].strip(),
                    glider_id=fields[13].strip(),
                    _igc_lambda=igc_lambda(self, index)))
            line = self.readline(0.5)
        return add_igc_filenames(tracks, self.manufacturer[:3].upper(), self.serial_number)

    def iact21(self, index):
        self.write('ACT_21_%02X\r\n' % index)
        while True:
            line = self.readline()
            yield line
            if line.startswith('G'):
                break

    def act22(self, index):
        self.write('ACT_22_00\r\n')
        line = self.readline()
        if line == 'ACT_22_00 Done\r\n':
            return True
        elif line == 'ACT_22_00 Fail\r\n':
            return False
        else:
            raise ProtocolError('unexpected response %r' % line)

    def act30(self):
        self.write('ACT_30_00\r\n')
        line = self.readline(2)
        if line != ' Done\r\n':
            raise ProtocolError('unexpected response %r' % line)

    def iact31(self):
        self.write('ACT_31_00\r\n')
        line = self.readline()
        if line == 'No Data\r\n':
            return
        while True:
            if line == ' Done\r\n':
                break
            m = re.match(r'\A(.*?);([NS])\s+(\d+)\'(\d+\.\d+);([EW])\s+(\d+)\'(\d+\.\d+);\s*(\d+);\s*(\d+)\r\n', line)
            if m:
                name = m.group(1)
                lat = int(m.group(3)) + float(m.group(4)) / 60.0
                if m.group(2) == 'S':
                    lat = -lat
                lon = int(m.group(6)) + float(m.group(7)) / 60.0
                if m.group(5) == 'W':
                    lon = -lon
                alt = int(m.group(8))
                radius = int(m.group(9))
                yield Waypoint(name, lat, lon, alt, radius=radius)
            else:
                raise ProtocolError('unexpected response %r' % line)
            line = self.readline()

    def act31(self):
        return list(self.iact31())

    def act32(self, waypoint):
        self.write('ACT_32_00\r\n')
        name = INVALID_CHARS_RE.sub('', waypoint.get_id_name())[:16].ljust(16)
        lat_hemi = 'N' if waypoint.lat > 0 else 'S'
        lat_deg, lat_min = divmod(abs(60 * waypoint.lat), 60)
        lon_hemi = 'E' if waypoint.lon > 0 else 'W'
        lon_deg, lon_min = divmod(abs(60 * waypoint.lon), 60)
        self.write('%s;%s  %2d\'%6.3f;%s %3d\'%6.3f;%6d;%6d\r\n' % (
            name,
            lat_hemi, lat_deg, lat_min,
            lon_hemi, lon_deg, lon_min,
            waypoint.alt or 0,
            waypoint.radius or 400))
        line = self.readline()
        if line == ' Done\r\n':
            return name
        elif line == 'full list\r\n':
            raise RuntimeError # FIXME
        elif line == 'Syntax Error\r\n':
            raise ProtocolError('syntax error')
        elif line == 'already exist\r\n':
            return name

    def act82(self):
        self.write('ACT_82_00\r\n')
        line = self.readline(0.1)
        if line != ' Done\r\n':
            raise ProtocolError('unexpected response %r' % line)

    def actbd(self):
        self.write('ACT_BD_00\r\n')
        return self.readline().strip()

    def rxa(self, x, parameter, format):
        self.write('R%cA_%02X\r\n' % (x, parameter))
        line = self.readline(0.2)
        m = re.match(r'\AR%cA_%02X_((?:[0-9A-F]{2})*)\r\n\Z' % (x, parameter), line)
        if m:
            return struct.unpack(format, ''.join(chr(int(x, 16)) for x in re.findall(r'..', m.group(1))))
        elif line == 'No Par\r\n':
            return None
        else:
            raise ProtocolError('unexpected response %r' % line)

    def rfa(self, parameter):
        return self.rxa('F', parameter, FA_FORMAT[parameter])

    def rpa(self, parameter):
        return self.rxa('P', parameter, PA_FORMAT[parameter])

    def wfa(self, parameter, value):
        format = FA_FORMAT[parameter]
        m = re.match(r'(\d+)s\Z', format)
        if m:
            width = int(m.group(1))
            value = INVALID_CHARS_RE.sub('', value)[:width].ljust(width)
        command = 'WFA_%02X_%s\r\n' % (parameter, ''.join('%02X' % ord(c) for c in struct.pack(format, value)))
        self.write(command)
        line = self.readline(0.1)
        if line == command:
            pass
        elif line == 'No Par\r\n':
            raise ProtocolError('there are no data defined for this parameter number')
        elif line == 'not ready\r\n':
            raise ProtocolError('missing the action $82')
        else:
            raise ProtocolError('Unexpected response %r' % line)

    def read_flash_response(self, expected='S003Done16', timeout=1):
        while True:
            m = FLASH_RESPONSE_RE.match(self.buffer)
            if m:
                response = m.group(1)
                self.buffer = self.buffer[len(m.group(1)):]
                logger.info('read %r' % response)
                if expected and response != expected:
                    raise FlashError('expected %r, got %r' % (expected, response))
                return response
            else:
                data = self.io.read(timeout)
                if len(data) == 0:
                    raise ReadError
                self.buffer += data

    def flash(self, model, srf):
        if model != '6015':
            raise RuntimeError # FIXME
        while True:
            self.write('S007FlyProg_6015AA')
            response = self.read_flash_response(None)
            if response == 'S007Flash6015_V1F8':
                break
            elif response == 'S0056015Wait39':
                time.sleep(1)
            else:
                raise FlashError('unexpected response %r' % response)
        self.write('S004Erase_4C')
        self.read_flash_response('S004Erase_4C')
        self.read_flash_response()
        for i, record in enumerate(srf.records):
            self.write(record)
            self.read_flash_response()
            yield (i + 1, len(srf.records))

    @property
    def manufacturer(self):
        if self._bd is None:
            self._bd = self.actbd()
        return self.ATTRIBUTES[self._bd]['manufacturer']

    @property
    def model(self):
        if self._bd is None:
            self._bd = self.actbd()
        return self.ATTRIBUTES[self._bd]['model']

    @property
    def serial_number(self):
        if self._serial_number is None:
            self._serial_number = self.rpa(PA_DeviceNr)[0]
        return self._serial_number

    @property
    def software_version(self):
        if self._software_version is None:
            value = self.rpa(PA_SoftVers)[0]
            self._software_version = '%d.%d.%02d' % (value / 1000, (value / 100) % 10, value % 100)
        return self._software_version

    @property
    def pilot_name(self):
        if self._pilot_name is None:
            self._pilot_name = self.rfa(FA_Owner)[0].strip()
        return self._pilot_name

    def get(self, key):
        if not key in FA_MAP:
            raise NotAvailableError
        return self.rfa(FA_MAP[key])[0]

    def set(self, key, value, first=True, last=True):
        if not key in FA_MAP:
            raise NotAvailableError
        if first:
            self.act82()
        if key == FA_Owner:
            self._pilot_name = None
        self.wfa(FA_MAP[key], value)

    def tracks(self):
        if self._tracks is None:
            self._tracks = self.act20()
        return self._tracks

    def waypoints(self):
        return self.iact31()

    def waypoint_remove(self, name=None):
        if name:
            raise NotAvailableError
        else:
            self.act30()

    def waypoint_upload(self, waypoint):
        return self.act32(waypoint)

    def to_json(self):
        fa = dict((key, self.rfa(key)) for key in FA_FORMAT.keys())
        pa = dict((key, self.rpa(key)) for key in PA_FORMAT.keys())
        tracks = list(track.to_json(True) for track in self.tracks())
        waypoints = list(waypoint.to_json() for waypoint in self.waypoints())
        return dict(fa=fa, pa=pa, tracks=tracks, waypoints=waypoints)



#   common.py  Flytec and Brauniger common functions
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re


def simplerepr(obj):
    keys = sorted(key for key in obj.__dict__.keys() if not key.startswith('_'))
    attrs = ''.join(' %s=%r' % (key, obj.__dict__[key]) for key in keys)
    return '<%s%s>' % (obj.__class__.__name__, attrs)


class CTR(object):

    def __init__(self, name, warning_distance, remark, ctrpoints):
        self.name = name
        self.warning_distance = warning_distance
        self.remark = remark
        self.ctrpoints = ctrpoints

    def to_json(self):
        json = {}
        json['name'] = self.name
        json['warning_distance'] = self.warning_distance
        json['remark'] = self.remark
        json['points'] = [ctrpoint.to_json() for ctrpoint in self.ctrpoints]
        return json

    __repr__ = simplerepr


class CTRPoint(object):

    def __init__(self, type, lat, lon, radius=None, clockwise=None):
        self.type = type
        self.lat = lat
        self.lon = lon
        self.radius = radius
        self.clockwise = clockwise

    def to_json(self):
        json = {}
        json['type'] = self.type
        json['lat'] = self.lat
        json['lon'] = self.lon
        if self.type == 'C':
            json['radius'] = self.radius
        elif self.type in ('T', 'Z'):
            json['clockwise'] = self.clockwise
        return json

    __repr__ = simplerepr


class Track(object):

    def __init__(self, **kwargs):
        self._igc = None
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def igc(self):
        if self._igc is None:
            self._igc = []
            for line in self._igc_lambda():
                yield line
                self._igc.append(line)
        else:
            for line in self._igc:
                yield line

    def to_json(self, igc=False):
        json = {}
        for key, value in self.__dict__.items():
            if key.startswith('_'):
                continue
            elif key == 'datetime':
                value = value.strftime('%Y-%m-%dT%H:%M:%SZ')
            elif key == 'duration':
                minutes, seconds = divmod(value.seconds, 60)
                hours, minutes = divmod(minutes, 60)
                value = '%02d:%02d:%02d' % (hours, minutes, seconds)
            json[key] = value
        if igc:
            json['igc'] = list(self.igc)
        return json


def add_igc_filenames(tracks, manufacturer, serial_number):
    date, index = None, 0
    for track in reversed(tracks):
        if track.datetime.date() == date:
            index += 1
        else:
            index = 1
        track.igc_filename = '%s-%s-%d-%02d.IGC' % (track.datetime.strftime('%Y-%m-%d'), manufacturer, serial_number, index)
        date = track.datetime.date()
    return tracks


def parse_openair(io):
    ctr = None
    for l in io:
        l = l.strip()
        if l.startswith('*'):
            continue
        if l.startswith('AC '):
            ctr = CTR(None, None, None, [])
            ac = l[3:].strip()
        elif l.startswith('AN '):
            ctr.name = l[3:].strip()
        elif l.startswith('AL '):
            al = l[3:].strip()
        elif l.startswith('AH '):
            ah = l[3:].strip()
            ctr.remark = '%s %s-%s' % (ac, al, ah)
        elif l.startswith('DP '):
            m = re.match(r'\ADP\s+(\d+):(\d+):(\d+)\s+([NS])\s+(\d+):(\d+):(\d+)\s+([EW])\Z', l)
            assert m is not None
            lat = int(m.group(1)) + int(m.group(2)) / 60.0 + int(m.group(3)) / 3600.0
            if m.group(4) == 'S':
                lat = -lat
            lon = int(m.group(5)) + int(m.group(6)) / 60.0 + int(m.group(7)) / 3600.0
            if m.group(8) == 'S':
                lon = -lon
            ctrpoint = CTRPoint('P', lat, lon)
            ctr.ctrpoints.append(ctrpoint)
        elif not l:
            if ctr:
                yield ctr
            ctr = None
        else:
            print repr(l)  # FIXME
    if ctr:
        yield ctr



#   flightrecorder.py  Flight recorder generic functions
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


from glob import glob
import logging
import re
import os

from errors import TimeoutError
from fifty20 import Fifty20
from flymaster import Flymaster
from serialio import SerialIO
from sixty15 import Sixty15


logger = logging.getLogger(__name__)


DEVICE_GLOBS = {
        'Darwin': (
            '/dev/cu.PL2303*',
            '/dev/cu.usbserial*',),
        'FreeBSD': (
            '/dev/cuad*',),
        'Linux': (
            '/dev/ttyUSB*',)}


class FlightRecorder(object):

    SUPPORTED_MODELS = Fifty20.SUPPORTED_MODELS + Flymaster.SUPPORTED_MODELS + Sixty15.SUPPORTED_MODELS

    def __new__(self, device=None, model=None):
        if device:
            devices = (device,)
        else:
            device_globs = DEVICE_GLOBS.get(os.uname()[0], ())
            devices = list(filename for device_glob in device_globs for filename in sorted(glob(device_glob)))
        if model is not None and model not in FlightRecorder.SUPPORTED_MODELS:
            raise RuntimeError # FIXME
        for device in devices:
            try:
                io = SerialIO(device)
            except IOError:
                continue
            if model in Fifty20.SUPPORTED_MODELS:
                return Fifty20(io)
            elif model in Flymaster.SUPPORTED_MODELS:
                return Flymaster(io)
            elif model in Sixty15.SUPPORTED_MODELS:
                return Sixty15(io)
            elif model is None:
                try:
                    try:
                        line = 'PBRSNP,'.encode('nmea_sentence')
                        logger.info('write %r' % line)
                        io.write(line)
                        line = io.read(0.2)
                        while line.find('\x11' if line[0] == '\x13' else '\n') == -1:
                            line += io.read()
                        logger.info('readline %r' % line)
                        if re.match('\x13\$PBRSNP,[^,]*,[^,]*,[^,]*,[^,]*\*[0-9A-F]{2}\r\n\x11\Z', line):
                            return Fifty20(io, line)
                        if re.match('\$PBRSNP,[^,]*,[^,]*,[^,]*,[^,]*,[^,]*,[^,]*\*[0-9A-F]{2}\r\n\Z', line):
                            return Flymaster(io, line)
                    except TimeoutError:
                        line = 'ACT_BD_00\r\n'
                        logger.info('write %r' % line)
                        io.write(line)
                        line = io.read(0.2)
                        while line.find('\n') == -1:
                            line += io.read()
                        logger.info('readline %r' % line)
                        if re.match('(Brauniger IQ-Basic|Flytec 6015)\r\n\Z', line):
                            logger.info('read %r' % line)
                            return Sixty15(io, line)
                except TimeoutError:
                    pass
        raise TimeoutError



#    waypoint.py  Waypoint functions
#    Copyright (C) 2011  Tom Payne
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import re


logger = logging.getLogger(__name__)


class WaypointError(RuntimeError):
    pass


class Waypoint(object):

    def __init__(self, name, lat, lon, alt, airfield=None, color=None, id=None, radius=None):
        id2, alt2, airfield2 = None, None, None
        if id is not None:
            m = re.match('([A-Z]\d{2})(\d{3})?', id)
            if m:
                id2 = m.group(1)
                if m.group(2):
                    alt2 = 10 * int(m.group(2))
                if id.startswith('A'):
                    airfield2 = True
        id3, alt3, name3, airfield3 = None, None, None, None
        if name is not None:
            m = re.match('([A-Z]\d{2})(\d{3})?(?:\s+(.*))?\Z', name)
            if m:
                id3 = m.group(1)
                if id3.startswith('A'):
                    airfield3 = True
                if m.group(2):
                    alt3 = 10 * int(m.group(2))
                name3 = m.group(3) or ''
                if re.search('attero|goal|land', name, re.I):
                    airfield3 = True
        for n in (name3, name):
            if n is not None:
                self.name = n.rstrip()
                break
        else:
            self.name = ''
        for i in (id2, id3, id):
            if i is not None:
                self.id = i
                break
        else:
            self.id = ''
        self.lat = lat
        self.lon = lon
        for a in (alt, alt2, alt3):
            if a is not None:
                self.alt = a
                break
        else:
            self.alt = None
        for a in (airfield, airfield2, airfield3):
            if a is not None:
                self.airfield = a
                break
        else:
            self.airfield = False
        self.color = color
        self.radius = radius
        self.device_name = name

    def get_id(self):
        if re.match('[A-Z]\d{2}\Z', self.id):
            return '%s%s' % (self.id, '%03d' % (self.alt / 10) if self.alt else '')
        else:
            return self.id

    def get_id_name(self):
        if re.match('[A-Z]\d{2}\Z', self.id):
            return '%s%s%s' % (self.id, '%03d' % (self.alt / 10) if self.alt else '', ' %s' % self.name if self.name else '')
        else:
            return '%s%s' % (self.id, ' %s' % self.name if self.name else '')

    def to_json(self):
        return self.__dict__.copy()


def dump(waypoints, file, format='formatgeo'):
    if format == 'compegps':
        file.write(u'G  WGS 84\r\n')
        file.write(u'U  1\r\n')
        for waypoint in waypoints:
            file.write((u'W  %6s A %.10f\u00ba%s %.10f\u00ba%s 27-MAR-62 00:00:00 %f %s\r\n' % (
                    waypoint.get_id(),
                    abs(waypoint.lat),
                    'S' if waypoint.lat < 0 else 'N',
                    abs(waypoint.lon),
                    'W' if waypoint.lon < 0 else 'E',
                    waypoint.alt or -9999.0,
                    waypoint.name)).encode('iso-8859-1'))
            color = None if waypoint.color is None else int(waypoint.color[1:], 16)
            file.write(u'w Waypoint,0,-1.0,16777215,%s,1,7,%s\r\n' % (
                    '' if color is None else str(((color & 0xff) << 16) + (color & 0xff00) + (color >> 16)),
                    '' if waypoint.radius is None else str(waypoint.radius)))
    elif format == 'formatgeo':
        file.write(u'$FormatGEO\r\n')
        for waypoint in waypoints:
            file.write(u'%-6s    %s %02d %02d %05.2f    %s %03d %02d %05.2f  %4d  %s\r\n' % (
                    waypoint.get_id(),
                    'S' if waypoint.lat < 0 else 'N',
                    abs(waypoint.lat),
                    (60 * abs(waypoint.lat)) % 60,
                    (3600 * abs(waypoint.lat)) % 60,
                    'W' if waypoint.lon < 0 else 'E',
                    abs(waypoint.lon),
                    (60 * abs(waypoint.lon)) % 60,
                    (3600 * abs(waypoint.lon)) % 60,
                    waypoint.alt or 0,
                    waypoint.name))
    elif format == 'oziexplorer':
        file.write(u'OziExplorer Waypoint File Version 1.0\r\n')
        file.write(u'WGS 84\r\n')
        file.write(u'Reserved 2\r\n')
        file.write(u'Reserved 3\r\n')
        for i, waypoint in enumerate(waypoints):
            color = None if waypoint.color is None else int(waypoint.color[1:], 16)
            file.write(u'%d,%s,%f,%f,,,1,,%s,,%s,,,%s,%s\r\n' % (
                    i + 1,
                    waypoint.get_id(),
                    waypoint.lat,
                    waypoint.lon,
                    '%d' % (((color & 0xff) << 16) + (color & 0xff00) + (color >> 16)) if color is not None else '',
                    waypoint.name,
                    '' if waypoint.radius is None else str(waypoint.radius),
                    '-777' if waypoint.alt is None else str(waypoint.alt / 0.3048)))
    elif format == 'seeyou':
        file.write(u'title,code,country,latitude,longitude,elevation,style,direction,length,frequency,description\r\n')
        for waypoint in waypoints:
            file.write(u'"%s","%s",,%02d%06.3f%s,%03d%06.3f%s,%s,,,,,\r\n' % (
                    waypoint.name,
                    waypoint.get_id(),
                    abs(waypoint.lat),
                    (60 * abs(waypoint.lat)) % 60,
                    'S' if waypoint.lat < 0 else 'N',
                    abs(waypoint.lon),
                    (60 * abs(waypoint.lon)) % 60,
                    'W' if waypoint.lon < 0 else 'E',
                    '' if waypoint.alt is None else '%fm' % waypoint.alt))


def load(fp, encoding='iso-8859-1'):
    lines = list(line.rstrip() for line in fp.read().decode(encoding).splitlines())
    projs = {}
    waypoints = []
    # FIXME horrible hack to remove byte order mark and new B line from CompeGPS files
    if len(lines) >= 1 and re.search(r'B\s+UTF-8\Z', lines[0]):
        lines = lines[1:]
    if len(lines) >= 2 and re.match(r'\AG\s+WGS\s+84\s*\Z', lines[0]) and re.match(r'\AU\s+1\s*\Z', lines[1]):
        for line in lines[2:]:
            if not line:
                continue
            m = re.match(r'\AW\s+(\S+)\s+A\s+(\d+\.\d+).*([NS])\s+(\d+\.\d+).*([EW])\s+\S+\s+\S+\s+(-?\d+(?:\.\d+))(?:\s+(.*))?\Z', line)
            if m:
                id = m.group(1)
                lat = float(m.group(2))
                if m.group(3) == 'S':
                    lat = -lat
                lon = float(m.group(4))
                if m.group(5) == 'W':
                    lon = -lon
                alt = float(m.group(6))
                name = m.group(7) or ''
                waypoints.append(Waypoint(name, lat, lon, alt if alt > 0 else None, id=id))
                continue
            m = re.match(r'\AW\s+(\S+)\s+(\d+)([CDEFGHJKLMNPQRSTUVWX])\s+(\d+)\s+(\d+)\s+\d{2}-(?:JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC)-\d{2}\s+\d{2}:\d{2}:\d{2}\s+(-?\d+(?:\.\d+))(?:\s+(.*))?\Z', line)
            if m:
                from pyproj import Proj
                id = m.group(1)
                zone = m.group(2)
                if zone not in projs:
                    projs[zone] = Proj(proj='utm', zone=zone, ellps='WGS84')
                x, y = int(m.group(4)), int(m.group(5))
                if m.group(3) < 'N':
                    y -= 10000000
                lon, lat = projs[zone](x, y, inverse=True)
                name = m.group(7)
                waypoints.append(Waypoint(name, lat, lon, alt if alt > 0 else None, id=id))
                continue
            m = re.match('\Aw\s+[^,]*,[^,]*,-?\d+(?:\.\d+)?,[^,]*,(\d*),[^,]*,[^,]*,[^,]*(?:,(-?\d+(?:\.\d+)?))?', line)
            if m and len(waypoints) > 0:
                waypoint = waypoints[-1]
                if m.group(1):
                    color = int(m.group(1))
                    waypoint.color = '#%02x%02x%02x' % (color & 0xff, (color >> 8) & 0xff, (color >> 16) & 0xff)
                if m.group(2):
                    waypoint.radius = float(m.group(2))
                continue
            m = re.match(r'\Az', line)
            if m:
                continue
            logger.warning('unrecognized waypoint %r' % line)
    elif len(lines) >= 1 and re.match(r'\A\$FormatGEO\s*\Z', lines[0]):
        for line in lines[1:]:
            if not line:
                continue
            m = re.match(r'\A(\S+)\s+([NS])\s+(\d+)\s+(\d+)\s+(\d+\.\d+)\s+([EW])\s+(\d+)\s+(\d+)\s+(\d+\.\d+)\s+(-?\d+)(?:\s+(.*))?\Z', line)
            if m:
                id = m.group(1)
                lon = int(m.group(7)) + int(m.group(8)) / 60.0 + float(m.group(9)) / 3600.0
                if m.group(2) == 'S':
                    lat = -lat
                lat = int(m.group(3)) + int(m.group(4)) / 60.0 + float(m.group(5)) / 3600.0
                if m.group(6) == 'W':
                    lon = -lon
                alt = int(m.group(10))
                name = m.group(11)
                waypoints.append(Waypoint(name, lat, lon, alt, id=id))
                continue
            logger.warning('unrecognized waypoint %r' % line)
    elif len(lines) >= 1 and re.match(r'\A\$FormatUTM\s*\Z', lines[0]):
        for line in lines[1:]:
            if not line:
                continue
            m = re.match(r'\A(\S+)\s+(\d+)([A-Z])\s+(\d+)\s+(\d+)\s+(-?\d+)(?:\s+(.*))?\Z', line)
            if m:
                from pyproj import Proj
                id = m.group(1)
                zone = m.group(2)
                if zone not in projs:
                    projs[zone] = Proj(proj='utm', zone=zone, ellps='WGS84')
                x, y = int(m.group(4)), int(m.group(5))
                if m.group(3) < 'N':
                    y -= 10000000
                lon, lat = projs[zone](x, y, inverse=True)
                alt = int(m.group(6))
                name = m.group(7)
                waypoints.append(Waypoint(name, lat, lon, alt, id=id))
                continue
            logger.warning('unrecognized waypoint %r' % line)
    elif len(lines) > 1 and re.match(r'\Atitle,code,country,latitude,longitude,elevation,style,direction,length,frequency,description', lines[0], re.I):
        columns = re.split(r'\s*,\s*', lines[0].rstrip().lower())
        for line in lines[1:]:
            if not line:
                continue
            try:
                fields = dict((columns[i], value) for (i, value) in enumerate(re.split(r'\s*,\s*', line.rstrip())))
                m = re.match(r'\A(\d\d)(\d\d\.\d\d\d)([NS])\Z', fields['latitude'])
                if not m:
                    raise WaypointError
                lat = int(m.group(1)) + float(m.group(2)) / 60.0
                if m.group(3) == 'S':
                    lat = -lat
                m = re.match(r'\A(\d\d\d)(\d\d\.\d\d\d)([EW])\Z', fields['longitude'])
                if not m:
                    raise WaypointError
                lon = int(m.group(1)) + float(m.group(2)) / 60.0
                if m.group(3) == 'W':
                    lon = -lon
                m = re.match(r'\A(\d+(?:\.\d*)?)(m|ft)\Z', fields['elevation'])
                if not m:
                    raise WaypointError
                alt = float(m.group(1))
                if m.group(2) == 'ft':
                    alt *= 0.3048
                id = fields['code']
                if id[0] == '"' and id[-1] == '"':
                    id = id[1:-1]
                name = fields['title']
                if name[0] == '"' and name[-1] == '"':
                    name = name[1:-1]
                waypoints.append(Waypoint(name, lat, lon, alt, id=id))
            except WaypointError:
                logger.warning('unrecognized waypoint %r' % line)
    elif len(lines) >= 4 and re.match(r'\AOziExplorer\s+Waypoint\s+File\s+Version\s+\d+\.\d+\Z', lines[0]) and re.match(r'\AWGS\s+84\s*\Z', lines[1]):
        for line in lines[4:]:
            if not line:
                continue
            fields = re.split(r'\s*,\s*', line)
            id = fields[1]
            lat = float(fields[2])
            lon = float(fields[3])
            if fields[9]:
                color = int(fields[9])
                color = '#%02x%02x%02x' % (color & 0xff, (color >> 8) & 0xff, (color >> 16) & 0xff)
            else:
                color = None
            name = re.sub(r'\xd1', ',', fields[10])
            if len(fields) > 13 and fields[13] and float(fields[13]) > 0.0:
                radius = float(fields[13])
            else:
                radius = None
            alt = 0.3048 * float(fields[14]) if fields[14] != '-777' else None
            waypoints.append(Waypoint(name, lat, lon, alt, color=color, id=id, radius=radius))
    else:
        logger.error('unrecognised waypoint format %r' % lines[0])
    return waypoints


if __name__ == '__main__':
    import sys
    dump(load(sys.stdin), sys.stdout, format='compegps')



#   serialio.py  Flytec and Brauniger serial I/O functions
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import os
import select
import termios
import time
import tty


from errors import TimeoutError, WriteError


logger = logging.getLogger(__name__)


class SerialIO(object):

    def __init__(self, filename, speed=tty.B57600):
        try:
            self.filename = filename
            logger.info('opening %r' % filename)
            self.fd = os.open(filename, os.O_RDWR | os.O_NOCTTY | os.O_NONBLOCK)
            tty.setraw(self.fd)
            attr = tty.tcgetattr(self.fd)
            attr[tty.ISPEED] = attr[tty.OSPEED] = speed
            tty.tcsetattr(self.fd, tty.TCSAFLUSH, attr)
        except termios.error:
            raise IOError

    def set_speed(self, speed):
        attr = tty.tcgetattr(self.fd)
        attr[tty.ISPEED] = attr[tty.OSPEED] = speed
        tty.tcsetattr(self.fd, tty.TCSAFLUSH, attr)

    def read(self, timeout=1, n=1024):
        if select.select([self.fd], [], [], timeout) == ([], [], []):
            raise TimeoutError
        data = os.read(self.fd, n)
        logger.debug('%.3f read %r (%d bytes)' % (time.time(), data, len(data)))
        return data

    def readn(self, n, timeout=1):
        data = ''
        while len(data) < n:
            data += self.read(timeout, n - len(data))
        return data

    def write(self, line):
        logger.debug('%.3f write %r (%d bytes)' % (time.time(), line, len(line)))
        if os.write(self.fd, line) != len(line):
            raise WriteError

    def flush(self):
        tty.tcflush(self.fd, tty.TCIOFLUSH)



#   nmea.py  NMEA codec
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


from codecs import Codec, CodecInfo
import codecs
import re


NMEA_ENCODE_RE = re.compile('\\A[\x20-\x7e]{1,79}\\Z')
NMEA_DECODE_RE = re.compile('\\A\\$(.{1,79})\\*([0-9A-F]{2})\r\n\\Z')
NMEA_INVALID_CHAR_RE = re.compile('[^\x20-\x7e]+')


class NMEAError(UnicodeError):
    pass


class NMEASentenceCodec(Codec):

    def decode(self, input, errors='strict'):
        if errors != 'strict':
            raise NotImplementedError
        if not input:
            return ('', 0)
        m = NMEA_DECODE_RE.match(input)
        if not m:
            raise NMEAError(input)
        checksum = 0
        for c in m.group(1):
            checksum ^= ord(c)
        if checksum != ord(m.group(2).decode('hex')):
            raise NMEAError(input)
        return (m.group(1), len(input))

    def encode(self, input, errors='strict'):
        if errors != 'strict':
            raise NotImplementedError
        if not input:
            return ('', 0)
        if not NMEA_ENCODE_RE.match(input):
            raise NMEAError(input)
        checksum = 0
        for c in input:
            checksum ^= ord(c)
        return ('$%s*%02X\r\n' % (input, checksum), len(input))


class NMEACharacterCodec(object):

    def decode(self, input, errors='strict'):
        return (unicode(input), len(input))

    def encode(self, input, errors='strict'):
        if errors == 'replace':
            return (NMEA_INVALID_CHAR_RE.sub(lambda m: '?' * len(m.group()), input), len(input))
        elif errors == 'strict':
            if NMEA_INVALID_CHAR_RE.search(input):
                raise UnicodeError
            return (input, len(input))
        else:
            raise NotImplementedError


def nmea_search(encoding):
    if encoding == 'nmea_sentence':
        codec = NMEASentenceCodec()
        return CodecInfo(codec.encode, codec.decode, name=encoding)
    if encoding == 'nmea_characters':
        codec = NMEACharacterCodec()
        return CodecInfo(codec.encode, None, name=encoding)
    return None


codecs.register(nmea_search)



#   firmware.py  Flytec and Brauniger firmware functions
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import re
import select
import string
import struct
from itertools import cycle, izip
import zipfile


VIGENERE_ALPHABET = string.ascii_uppercase + string.ascii_lowercase + string.digits
VIGENERE_KEY = 'Verna12Mry34st'


class VigenereError(RuntimeError):
    pass


class Vigenere(object):

    def __init__(self, alphabet, key):
        self.alphabet = alphabet
        self.indices = dict((v, i) for i, v in enumerate(self.alphabet))
        self.key = key
        self.reset()

    def encode(self, s):
        try:
            return ''.join(self.alphabet[(self.indices[c] + self.indices[k]) % len(self.alphabet)] for c, k in izip(s, self.ikey))
        except KeyError:
            raise VigenereError

    def decode(self, s):
        try:
            return ''.join(self.alphabet[(self.indices[c] - self.indices[k]) % len(self.alphabet)] for c, k in izip(s, self.ikey))
        except KeyError:
            raise VigenereError

    def reset(self):
        self.ikey = cycle(self.key)


S0_RE = re.compile(r'S0([0-9A-F]{2})0000((?:[0-9A-F]{2})+)([0-9A-F]{2})\Z')
S1_RE = re.compile(r'S(1)([0-9A-F]{2})([0-9A-F]{4})((?:[0-9A-F]{2})+)([0-9A-F]{2})\Z')
S2_RE = re.compile(r'S(2)([0-9A-F]{2})([0-9A-F]{6})((?:[0-9A-F]{2})+)([0-9A-F]{2})\Z')
S3_RE = re.compile(r'S(3)([0-9A-F]{2})([0-9A-F]{8})((?:[0-9A-F]{2})+)([0-9A-F]{2})\Z')
S5_RE = re.compile(r'S503([0-9A-F]{4})([0-9A-F]{2})\Z')
S7_RE = re.compile(r'S705([0-9A-F]{8})([0-9A-F]{2})\Z')
S8_RE = re.compile(r'S804([0-9A-F]{6})([0-9A-F]{2})\Z')
S9_RE = re.compile(r'S903([0-9A-F]{4})([0-9A-F]{2})\Z')


class SRecordError(RuntimeError):

    def __init__(self, line):
        self.line = line


class SRecordFile(object):

    def __init__(self, lines):
        self.header = None
        self.data = {}
        self.starting_executing_address = None
        self.records = list(line.rstrip() for line in lines)
        for record in self.records:
            m = S0_RE.match(record)
            if m:
                length = int(m.group(1), 16)
                data = ''.join(chr(int(x, 16)) for x in re.findall(r'..', m.group(2)))
                checksum = int(m.group(3), 16) # FIXME check
                if length != len(data) + 3:
                    raise SRecordError(record)
                self.header = data
                continue
            m = S1_RE.match(record) or S2_RE.match(record) or S3_RE.match(record)
            if m:
                i = int(m.group(1))
                length = int(m.group(2), 16)
                address = int(m.group(3), 16)
                data = ''.join(chr(int(x, 16)) for x in re.findall(r'..', m.group(4)))
                checksum = int(m.group(5), 16)
                if length != 2 + i + len(data):
                    raise SRecordError(record)
                self.data[address] = data
                continue
            m = S5_RE.match(record)
            if m:
                address = int(m.group(1), 16)
                checksum = int(m.group(2), 16)
                if address != len(self.srecords):
                    raise SRecordError(record)
                continue
            m = S7_RE.match(record) or S8_RE.match(record) or S9_RE.match(record)
            if m:
                address = int(m.group(1), 16)
                checksum = int(m.group(2), 16)
                self.starting_executing_address = address
                continue
            raise SRecordError(record)

    def pages(self):
        data_offset, data, data_length = None, [], 0
        for address in sorted(self.data.keys()):
            if data_offset is None:
                data_offset = address & ~0xff
            fill_length = address - data_offset - data_length
            if fill_length > 0:
                data.append('\xff' * fill_length)
                data_length += fill_length
            elif fill_length != 0:
                raise SRecordError
            data.append(self.data[address])
            data_length += len(self.data[address])
        incomplete = data_length & 0xff
        if incomplete:
            data.append('\xff' * (size - incomplete))
            data_length += incomplete
        data = ''.join(data)
        for i in xrange(0, data_length >> 8):
            yield ((data_offset >> 8) + i, data[256 * i:256 * (i + 1)])


class M32C87Error(RuntimeError):

    def __init__(self, message=None):
        self.message = message


class M32C87(object):

    # Status register (SRD)
    WSM_READY = 0x80
    ERASE_ERROR = 0x20
    PROGRAM_ERROR = 0x10
    BLOCK_ERROR = 0x08

    # Status register 1 (SRD1)
    BOOT_UPDATE_COMPLETE = 0x80
    CHECKSUM_MATCH = 0x10
    DATA_RECEIVE_TIMEOUT = 0x02

    def __init__(self, io):
        self.io = io

    def sleep(self, timeout):
        select.select([], [], [], timeout)

    def initialize(self):
        for i in xrange(0, 16):
            self.io.write('\x00')
            self.sleep(0.02)
        self.sleep(0.05)
        if self.io.readn(1) != '\xb0':
            raise M32C87Error('initialize')

    def set_speed(self, speed):
        c = {tty.B9600: '\xb0', tty.B19200: '\xb1', tty.B38400: '\xb2', tty.B57600: '\xb3', tty.B115200: '\xb4'}[speed]
        self.io.write(c)
        self.io.set_speed(speed)
        self.sleep(0.05)
        if self.io.readn(1) != c:
            raise M32C87Error('set_speed')

    def command(self, value, format='', args=(), output=None):
        self.io.write(struct.pack('>B' + format, value, *args))
        if output:
            return struct.unpack('>' + output, self.io.readn(struct.calcsize(output)))

    def check_id(self):
        pass # FIXME

    def lock(self):
        self.command('\x75')

    def unlock(self):
        self.command('\x7a')

    def erase(self):
        self.command('\xa7', 'B', ('\xd0',))
        if status_register_check(M32C87.ERASE_ERROR):
            raise M32C87Error('erase')

    def status_register_read(self):
        return self.command('\x70', '', (), 'BB')

    def status_register_check(self, bits):
        while True:
            srd = self.status_register_read()
            if srd[0] & M32C87.WSM_READY:
                return bool(srd[0] & bits)
            self.sleep(0.05)

    def status_register_clear(self):
        self.command('\x50')
        self.sleep(0.05)

    def page_write(self, page, data):
        self.command('\x41', 'H256s', (page, data))
        self.sleep(0.05)
        if self.status_register_check(M32C87.PROGRAM_ERROR):
            raise M32C87Error('page_write')

    # FIXME delete member functions below this line

    def page_erase(self, page):
        self.command(0x20, 'HB', (page, 0xd0))

    def page_erase_all_unlocked(self):
        self.command('\xa7', 'B', ('\xd0',))

    def page_read(self, page):
        return self.command('\xff', 'H', (page,), '256s')[0]

    def page_lock_get(self, page):
        return self.command('\x71', 'H', (page,), 'B')[0]

    def page_lock_set(self, page, value):
        self.command('\x77', 'HB', (page, 0xd0))

    def check_data_get(self):
        return self.command('\xfd', '', (), 'H')[0]


def decode(file):
    vigenere = Vigenere(VIGENERE_ALPHABET, VIGENERE_KEY)
    for line in file:
        yield vigenere.decode(line.rstrip())


def firmware_model(filename):
    for pattern, model in (('5020|CTINO', '5020'), ('6015', '6015')):
        if re.search(pattern, filename, re.I):
            return model
    return '6020'


def firmware(file):
    try:
        zf = zipfile.ZipFile(file, 'r')
        for zi in zf.infolist():
            try:
                yield (firmware_model(zi.filename), SRecordFile(decode(zf.open(zi))))
            except (VigenereError, SRecordError):
                continue
    except zipfile.BadZipfile:
        pass
    file.seek(0)
    m = re.search(r'[%s]{18,}\s+([%s]+\s+){128,}' % (VIGENERE_ALPHABET, VIGENERE_ALPHABET), file.read(), re.M)
    if m:
        try:
            yield (firmware_model(file.name), SRecordFile(decode(m.group().splitlines())))
        except (VigenereError, SRecordError):
            pass
    try:
        file.seek(0)
        yield (firmware_model(file.name), SRecordFile(file))
    except SRecordError:
        pass


if __name__ == '__main__':
    import sys
    if False:
        for model, srf in firmware(open(sys.argv[1])):
            print model, srf.header, len(srf.data)
    if True:
        from serialio import SerialIO
        import tty
        model = '5020'
        logging.basicConfig(level=logging.DEBUG)
        io = SerialIO('/dev/ttyUSB0', tty.B9600)
        m32c87 = M32C87(io)
        m32c87.initialize()
        m32c87.set_speed(tty.B19200 if model == '5020' else tty.B57600)
        sys.exit()
        m32c87.unlock()
        m32c87.erase()
        for address, page in srf.pages():
            m32c87.page_write(address, page)
        m32c87.lock()



#   base.py  Flight recorder base class
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


from errors import NotAvailableError


class FlightRecorderBase(object):

    @property
    def manufacturer(self):
        raise NotAvailableError

    @property
    def model(self):
        raise NotAvailableError

    @property
    def serial_number(self):
        raise NotAvailableError

    @property
    def software_version(self):
        raise NotAvailableError

    @property
    def pilot_name(self):
        raise NotAvailableError

    def ctri(self):
        raise NotAvailableError

    def ctrs(self):
        raise NotAvailableError

    def ctr_upload(self, ctr, warning_distance):
        raise NotAvailableError

    def flash(self, model, srf):
        raise NotAvailableError

    def get(self, key):
        raise NotAvailableError

    def set(self, key, value, first=True, last=True):
        raise NotAvailableError

    def tracks(self):
        raise NotAvailableError

    def waypoints(self):
        raise NotAvailableError

    def waypoint_remove(self, name=None):
        raise NotAvailableError

    def waypoint_upload(self, waypoint):
        raise NotAvailableError

    def to_json(self):
        raise NotAvailableError



#!/usr/bin/env python

import os
from setuptools import setup


version = '0.9.1'

long_description = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

setup(
        author='Tom Payne',
        author_email='twpayne@gmail.com',
        description='Utilities for flight recorders',
        long_description=long_description,
        name='flightrecorder',
        packages=['flightrecorder'],
        scripts=['scripts/flightrecorder'],
        setup_requires=['nose'],
        #test_suite='tests',
        url='https://github.com/twpayne/flightrecorder',
        version=version,
        zip_safe=True)
